-- ** ARCHIMONDE **
-- The archimonde boss mod.
-------------------------------------------------------------------------------

local NayMods = NayMods
local Debug = NayMods.debug
local Icons = NayMods.constants.icons
local Timer = NayMods.timer
local Utils = NayMods.utils
local Transmission = NayMods.transmission
local Options = NayMods.options
local Storage = NayMods.storage

local table_sort_desc = Utils.TableSortDescending
local table_sort_asc = Utils.TableSortAscending


local mod = NayMods:NewBossMod("Archimonde", 1799)
mod:RegisterEnableMob(91331)

local selectedTryLabel = nil
local currentTryLabel = nil
local markAssignments = {}
local markAssignmentsHistory = {}
local markWaitingList = {}
local markOfTheLegionCount = 0
local markOfTheLegionPositions = {
  [1] = {
    name = "Melee Left",
    icon = Icons.star
  },
  [2] = {
    name = "Melee Right",
    icon = Icons.circle
  },
  [3] = {
    name = "Ranged Left",
    icon = Icons.diamond
  },
  [4] = {
    name = "Ranged Right",
    icon = Icons.triangle
  }
}
local markOfLegionSpellIds = {
  187050, -- This should be the right one, but i'm just including all of them, safety first!
  190399,
  188514,
  187051
}

-----------------------------LOCAL-HELPERS-------------------------------------

--[[
Chat message based on the given position.
]]
local function positionChatMessage(position)
  if not position then
    Debug:Warning("Tried to create a say message for unknown position")
    return nil
  end

  return string.format("%s %s %s", position.icon.chatPattern, position.name, position.icon.chatPattern)
end

--[[
Visual alert message based on the given position.
]]
local function positionAlertMessage(position)
  if not position then
    Debug:Warning("Tried to create a visual message for unknown position")
    return nil
  end

  return string.format("%s {D} - %s %s", position.icon.texturePattern, position.name, position.icon.texturePattern)
end

--[[
System message based on the given position.
]]
local function positionMessage(position)
  if not position then
    Debug:Warning("Tried to create a header for unknown position")
    return nil
  end

  return string.format("%s %s %s", position.icon.texturePattern, position.name, position.icon.texturePattern)
end

-----------------------------OUT-OF-COMBAT-SETUP-------------------------------

--[[
Stores the given setup into storage.
]]
function mod:storeSetup(setup)
  Storage.db.char.archimonde.markSetup = setup
end

--[[
Gets the current setup found in storage.
]]
function mod:getSetup()
  return Storage.db.char.archimonde.markSetup
end

--[[
Removes the player currently in the given group and position.
]]
function mod:removePosition(groupNumber, position)
  if self:getSetup()[groupNumber] and self:getSetup()[groupNumber][position] then
      self:getSetup()[groupNumber][position] = nil
  end
end

--[[
Assigns the given player to the given group and position.
Also removes him from any position he had before, if any.
]]
function mod:assignPosition(playerName, groupNumber, position)
  local currentGroup, currentPosition = self:getPlayerAssignment(playerName)

  if currentGroup and currentPosition then
    self:removePosition(currentGroup, currentPosition)
  end

  self:getSetup()[groupNumber][position] = playerName
end

--[[
Gets the currently assigned player to the given group and position.
]]
function mod:getAssignment(groupNumber, position)
  if self:getSetup()[groupNumber] and self:getSetup()[groupNumber][position] then
      return self:getSetup()[groupNumber][position]
  end

  return nil
end

--[[
Gets the group and position the given player is in now.
]]
function mod:getPlayerAssignment(playerName)
  for groupNumber, players in pairs(self:getSetup()) do
    for position, name in pairs(players) do
      if name == playerName then
        return groupNumber, position
      end
    end
  end
  return nil, nil
end

--[[
Checks if the given player is assigned.
]]
function mod:isAssigned(playerName)
  local currentGroup, currentPosition = self:getPlayerAssignment(playerName)

  return currentGroup ~= nil and currentPosition ~= nil
end

--[[
Returns a colored array of the raid.
The color is there to indicate on who is assigned and who is not.
]]
function mod:RaidColored()
  local alreadyAssigned = {}
  local notAssigned = {}
  local coloredRaid = {}

  local raid = self:GetRaid()

  for k,v in pairs(raid) do
    local isAssigned = self:isAssigned(v)
    if not isAssigned then
      coloredRaid[k] = Utils:Color(v, "FF3333")
    else
      coloredRaid[k] = v
    end
  end

  return coloredRaid
end

--[[
Create the options group for all the positions.
]]
function mod:createPositionOptions(optionIndex)
  local order = optionIndex

  local options = {}

  for i=1,4 do
    local group = {
      type = "group",
      guiInline = true,
      name = positionMessage(markOfTheLegionPositions[i]),
      order = order,
      args = {}
    }
    for j=1,5 do
      group.args["pos" .. i .. "_" .. j] = {
        type = "select",
        name = "Player " .. j,
        width = "half",
        values = function(info) return self:RaidColored() end,
        get = function(info) return self:getAssignment(i, j) end,
        set = function(info, key)
           if self:CheckIAmPromoted() then
             self:assignPosition(key, i, j)
           else
             self:SysMessage("You don't have the permission to change the setup. You have to be at least a raid assistant.")
           end
           Options:Refresh()
         end,
        order = i
      }
    end

    options["group_" .. i] = group
  end

  options["pos_spacer"] = {
    type = "description",
    name = "\n",
    order = order + 1
  }
  options["broadcast"] = {
    type = "execute",
    name = "Broadcast setup",
    func = function()
      if self:CheckIAmPromoted() then
        Transmission:Send("ARCHIMONDE_SETUP", { markSetup = self:getSetup() }, "RAID")
      else
        self:SysMessage("You don't have the permission to broadcast the setup. You have to be at least a raid assistant.")
      end
    end,
    order = order  + 2
  }

  return options
end

--[[
Callback when we receive a new group setup.
]]
function mod:onSetupReceived(key, data, dist, sender)
  self:storeSetup(data.markSetup)
  Options:Refresh()
  if self:isAssigned(self.playerName) then
    local myGroup, myPosition = self:getPlayerAssignment(self.playerName)
    local position = markOfTheLegionPositions[myGroup]
    local positionMsg = string.format("You're now in %s %s %s!", position.icon.texturePattern, position.name, position.icon.texturePattern)

    self:Message("ARCHI_NEW_SETUP", positionMsg, 5)

    self:SysMessage("Archimonde setup has been updated by %s!", sender)
    self:SysMessage(positionMsg)
  end
end

--[[
OnInit hook.
]]
function mod:OnInit()
  Transmission:Register("ARCHIMONDE_SETUP", function(...) self:onSetupReceived(...) end)
end

--[[
Setup the default config for the Archimonde Bossmod.
]]
function mod:SetupDefaultConfig()
  self.defaultConfig = {
    char = {
      archimonde = {
        markSetup = {
          [1] = {},
          [2] = {},
          [3] = {},
          [4] = {}
        }
      },
      visuals = {
        [187050] = { -- Mark of the legion alert
          xOffset = 0,
          yOffset = 200,
          fontSize = 32
        },
        ARCHI_NEW_SETUP = { -- Alert on new setup
          xOffset = 0,
          yOffset = 250
        },
      }
    }
  }

end

--[[
Returns the Ace3 options sub-table for Archimonde.
]]
function mod:GetOptions()
  local this = self

  local args = {
    markLegionHeader = {
      type = "header",
      name = "Mark of the Legion",
      order = 1
    },
    markLegionPositionGroup = {
      type = "group",
      guiInline = true,
      name = "Positions",
      order = 2,
      args = self:createPositionOptions(2)
    },
    markLegionMessageGroup = {
      type = "group",
      guiInline = true,
      name = "Alert message",
      order = 4,
      args = {
        markLegionMessageFontSelector = self
          :Message(187050, positionAlertMessage(markOfTheLegionPositions[1]), nil, true)
          :FontSelector(1),
        markLegionMessageFontSizer = self
          :Message(187050, positionAlertMessage(markOfTheLegionPositions[1]), nil, true)
          :FontSizer(2),
        markLegionMessageToggleAnchor = self
          :Message(187050, positionAlertMessage(markOfTheLegionPositions[1]), nil, true)
          :ToggleAnchorOption(3),
      }
    },
    markLegionHistoryGroup = {
      type = "group",
      guiInline = true,
      name = "Mark of the Legion History",
      order = 5,
      args = {
        historySelect = {
          type = "select",
          name = "Attempt",
          values = function()
            local tryLabels = {}
            for k, _ in pairs(this:GetMarkAssignmentHistory()) do
              tryLabels[k] = k
            end
            return tryLabels
          end,
          get = function(info) return currentTryLabel end,
          set = function(info, key) selectedTryLabel = key end
        },
        historyToRaid = {
          type = "execute",
          name = "Report",
          func = function()
            if currentTryLabel == nil then return end
            this:RaidChat("NayMods: Mark of the Legion Report for Attempt: %s", currentTryLabel)

            local attemptAssignments = this:GetMarkAssignmentHistory()[currentTryLabel] or {}

            for k, v in pairs(attemptAssignments) do
              this:RaidChat("Round #%s", k)
              for p, n in pairs(v) do
                local position = markOfTheLegionPositions[p]
                this:RaidChat("  %s >> %s", n, position.name)
              end
            end
          end
        }
      }
    }
  }

  return {
    name = self.name,
    type = "group",
    args = args
  }
end

--[[
OnStart Hook.
]]
function mod:OnStart()
  Debug:WriteLine("Archimonde module started!")
  currentTryLabel = date("%m/%d/%y %H:%M:%S")
  self:Subscribe("SPELL_AURA_APPLIED", "MarkOfTheLegionApplied")
end

--[[
OnEngage hook.
]]
function mod:OnEngage()
  self:ResetMarks()
end

--[[
OnWipe hook.
]]
function mod:OnWipe()
  self:ResetMarks()
end

--[[
OnStop hook.
]]
function mod:OnStop()
  Debug:WriteLine("Archimonde module stopped!")
  self:Unsubscribe("SPELL_AURA_APPLIED")
  selectedTryLabel = currentTryLabel
  Options:Refresh()
end

-----------------------------COMBAT-MODULE-------------------------------

--[[
MarkOfTheLegionApplied handler.
]]
function mod:MarkOfTheLegionApplied(args)

  local player = args.destName

  if Utils:ArrayContains(markOfLegionSpellIds, args.spellId) or args.spellName == "Mark of the Legion" then
    Debug:WriteLine("Mark ID = %s", args.spellId)
    Debug:WriteLine("Mark of the Legion applied on %s!", player)
    markOfTheLegionCount = markOfTheLegionCount + 1

    local playerAssigned = self:TryAndAssignMarkToPresetGroup(player)

    if not playerAssigned then
      Debug:WriteLine("The preset group for %s is already taken. Will assign later", player)
      self:AddToWaitingList(player)
    end

    if markOfTheLegionCount == 4 then
      Debug:WriteLine("4 Marks of the Legion were handed out.")
      self:AssignWaitingList()
    end

    -- In 15 seconds, just reset the chain.
    -- We aren't checking if we reached 4 marks, because it can cause a bug when a player has disconnected
    -- and comes back right in the middle of mark distribution. For this we'll just schedule a reset on every mark.
    Timer.After(15, function() self:ResetMarks() end)
  end
end

--[[
Assign a given player to a given mark-group.
]]
function mod:AssignMark(player, groupNumber)
  local position = markOfTheLegionPositions[groupNumber] or nil

  if not position then
    Debug:Warning("Tried to assign %s to an invalid position %s", player, groupNumber)
    return
  end

  markAssignments[groupNumber] = player

  self:SysMessage("Mark of the Legion: %s >> %s", player, position.name)
  self:SetRaidIcon(player, position.icon.iconId)

  if self:IsMe(player) then
    local markExpiresAt, markDuration = self:GetPlayerMarkInfo()

    self:Say(positionChatMessage(position))
    self:Message(187050, positionAlertMessage(position), markDuration)

    Timer.After(markDuration - 3.3, function() self:Say("3") end)
    Timer.After(markDuration - 2.3, function() self:Say("2") end)
    Timer.After(markDuration - 1.3, function() self:Say("1") end)
  end
end

--[[
Gets the mark information on me.
]]
function mod:GetPlayerMarkInfo()
  local markOfTheLegion = GetSpellInfo(187050) or "Mark of the Legion"
  local markExpiresAt = select(7, UnitDebuff("player", markOfTheLegion))
  local markDuration = markExpiresAt - GetTime()

  return markExpiresAt, markDuration
end

--[[
Adds a player to the waiting list.
]]
function mod:AddToWaitingList(player)
  table.insert(markWaitingList, player)
end

--[[
Assigns everyone in the waiting list based on spec priotrity.
]]
function mod:AssignWaitingList()
  for _, p in pairs(markWaitingList) do
    self:AssignMarkWithSpecPrio(p)
  end
end

--[[
Tries to assign the player to his preset group.
]]
function mod:TryAndAssignMarkToPresetGroup(player)
  local playerGotAssigned = false

  for groupNumber, players in pairs(self:getSetup()) do
    for _, p in pairs(players) do
      if p == player and not markAssignments[groupNumber] then
        self:AssignMark(player, groupNumber)
        playerGotAssigned = true
      end
    end
  end

  return playerGotAssigned
end

--[[
Find all the open positions.
]]
function mod:FindAllOpenPositions()
  local openPositions = {}

  for i=1,4 do
    if not markAssignments[i] then
      table.insert(openPositions, i)
    end
  end

  return openPositions
end

--[[
Assign the player to an open spot based on his spec.
]]
function mod:AssignMarkWithSpecPrio(player)
  local isMelee = Utils:IsMelee(player)
  local openPositions = self:FindAllOpenPositions()

  if(isMelee) then
    table_sort_asc(openPositions)
  else
    table_sort_desc(openPositions)
  end

  local groupNumber = openPositions[1]

  self:AssignMark(player, groupNumber)
end

--[[
Reset the marks.
]]
function mod:ResetMarks()
  Debug:WriteLine("Resetting...")
  self:AddToHistory(markAssignments)
  self:ClearRaidIcons(markAssignments)
  markOfTheLegionCount = 0
  markAssignments = {}
  markWaitingList = {}
end

--[[
Returns the mark assignments history.
]]
function mod:GetMarkAssignmentHistory()
  return markAssignmentsHistory
end

--[[
Adds the given assignments to the history.
]]
function mod:AddToHistory(assignments)
  if assignments == nil or Utils:TableEmpty(assignments) then
    return
  end

  if not markAssignmentsHistory[currentTryLabel] then
    markAssignmentsHistory[currentTryLabel] = {}
  end

  table.insert(markAssignmentsHistory[currentTryLabel], assignments)
  Debug:Dump(markAssignmentsHistory)
end
